from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework import serializers
from rest_framework.validators import UniqueValidator
from rest_framework.authtoken.models import Token


User = get_user_model()


class DynamicFieldsModelSerializer(serializers.ModelSerializer):
    """
    A ModelSerializer that takes an additional `fields` argument that
    controls which fields should be displayed.
    """

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        fields = kwargs.pop('fields', None)

        # Instantiate the superclass normally
        super().__init__(*args, **kwargs)

        if fields is not None:
            # Drop any fields that are not specified in the `fields` argument.
            allowed = set(fields)
            existing = set(self.fields)
            for field_name in existing - allowed:
                self.fields.pop(field_name)


class UserSerializer(DynamicFieldsModelSerializer):
    class Meta:
        model = User
        fields = settings.CKL_REST_AUTH.get('REGISTER_FIELDS') + ('id',)

    def validate_username(self, value):
        if 'username' not in settings.CKL_REST_AUTH.get('REGISTER_FIELDS'):
            return None

        if not value:
            raise serializers.ValidationError('This field is required.')
        if User.objects.filter(username=value):
            raise serializers.ValidationError('This username is already in use.')

        return value

    def validate_email(self, value):
        if 'email' not in settings.CKL_REST_AUTH.get('REGISTER_FIELDS'):
            return None

        if not value:
            raise serializers.ValidationError('This field is required.')
        if User.objects.filter(email=value):
            raise serializers.ValidationError('This email is already in use.')

        return value


def RegisterSerializerFactory(user_serializer=UserSerializer):
    class RegisterSerializer(user_serializer):
        password = serializers.CharField(required=True)

        class Meta(user_serializer.Meta):
            fields = user_serializer.Meta.fields + ('password',)

        def create(self, validated_data):
            user = User.objects.create_user(**validated_data)
            token = Token.objects.create(user=user)

            return user, token

    return RegisterSerializer


class LoginSerializer(DynamicFieldsModelSerializer):
    username = serializers.CharField(required=True)
    email = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    class Meta:
        model = User
        fields = (
            'username',
            'email',
            'password'
        )


class PasswordResetSerializer(serializers.Serializer):
    email = serializers.EmailField(required=True)



import json
from pydoc import locate

import requests
from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.forms import PasswordResetForm
from django.core.exceptions import ImproperlyConfigured
from django.http import JsonResponse
from django.shortcuts import redirect
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import AllowAny
from rest_framework.views import APIView

from cklauth import constants
from cklauth.models import SocialAccount
from .serializers import RegisterSerializerFactory, LoginSerializer, PasswordResetSerializer


User = get_user_model()


class AuthError(Exception):
    def __init__(self, message, status):
        self.message = message
        self.status = status


class AuthView(APIView):
    status_code = status.HTTP_200_OK
    permission_classes = (AllowAny, )

    def post(self, request):
        try:
            user, token = self.perform_action(request)
        except AuthError as error:
            return JsonResponse(
                {'non_field_errors': [error.message]},
                status=error.status
            )

        UserSerializer = locate(settings.CKL_REST_AUTH.get('USER_SERIALIZER'))
        return JsonResponse({
            'token': token.key,
            'user': UserSerializer(instance=user).data,
        }, status=self.status_code)

    def perform_action(self, request):
        raise NotImplementedError('The view should implement `perform_login` method')


class RegisterView(AuthView):
    status_code = status.HTTP_201_CREATED

    def perform_action(self, request):
        UserSerializer = locate(settings.CKL_REST_AUTH.get('USER_SERIALIZER'))
        RegisterSerializer = RegisterSerializerFactory(UserSerializer)

        serializer = RegisterSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        return serializer.save()


class LoginView(AuthView):
    def perform_action(self, request):
        fields = [settings.CKL_REST_AUTH['LOGIN_FIELD'], 'password']
        serializer = LoginSerializer(data=request.data, fields=fields)

        serializer.is_valid(raise_exception=True)
        login_field = serializer.validated_data[settings.CKL_REST_AUTH['LOGIN_FIELD']]
        password = serializer.validated_data['password']
        user = authenticate(username=login_field, password=password)

        if not user:
            raise AuthError(message='Wrong credentials.', status=status.HTTP_401_UNAUTHORIZED)

        token, _ = Token.objects.get_or_create(user=user)
        return user, token


class SocialAuthView(AuthView):
    def __init__(self, *args, **kwargs):
        platform_settings = settings.CKL_REST_AUTH.get(self.platform, {})
        if not platform_settings:
            raise ImproperlyConfigured(
                'Add {} CLIENT_ID and REDIRECT_URI to settings.'.format(self.platform)
            )

        self.CLIENT_ID = platform_settings.get('CLIENT_ID')
        self.CLIENT_SECRET = platform_settings.get('CLIENT_SECRET')
        self.REDIRECT_URI = platform_settings.get('REDIRECT_URI')
        self.USER_INFO_MAPPING = platform_settings.get('USER_INFO_MAPPING')
        self.AUTH_FIELD_GENERATOR = platform_settings.get('AUTH_FIELD_GENERATOR')

        super().__init__(*args, **kwargs)

    def validate_response(self, response):
        if response.status_code != status.HTTP_200_OK:
            raise AuthError(message='Bad token.', status=status.HTTP_400_BAD_REQUEST)

    def get_access_token(self, request):
        if request.data.get('access_token'):
            return request.data.get('access_token')

        if not request.data.get('code'):
            raise AuthError(message='Missing auth token.', status=status.HTTP_400_BAD_REQUEST)

        payload = {
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_SECRET,
            'grant_type': 'authorization_code',
            'redirect_uri': self.REDIRECT_URI,
            'code': request.data['code'],
        }

        response = requests.post(self.token_url, data=payload)

        self.validate_response(response)

        return response.json()['access_token']

    def create_user(self, user_info, extra_fields={}):
        register_info = {
            register_key: (
                provider_key(user_info)
                if callable(provider_key)
                else user_info.get(provider_key)
            )
            for register_key, provider_key in self.USER_INFO_MAPPING.items()
        }

        if self.AUTH_FIELD_GENERATOR:
            auth_field_generator = locate(self.AUTH_FIELD_GENERATOR)
            register_info[User.USERNAME_FIELD] = auth_field_generator(register_info)

        register_info.update(extra_fields)

        UserSerializer = locate(settings.CKL_REST_AUTH.get('USER_SERIALIZER'))
        serializer = UserSerializer(data=register_info)
        serializer.is_valid(raise_exception=True)

        return User.objects.create_user(**serializer.data)

    def perform_action(self, request):
        access_token = self.get_access_token(request)
        user_info = self.get_user_info(access_token)

        try:
            # email registered with social account
            social_account = SocialAccount.objects.get(user__email=user_info.get('email'))
            user = social_account.user
            if not getattr(social_account, self.social_account_field):
                setattr(social_account, self.social_account_field, user_info.get('id'))
                social_account.save()
        except SocialAccount.DoesNotExist:
            try:
                # email registered without social account
                user = User.objects.get(email=user_info.get('email'))
                raise AuthError(
                    message='Registered with email.',
                    status=status.HTTP_400_BAD_REQUEST
                )
            except User.DoesNotExist:
                # user and social account don't exist
                extra_fields = request.data.get('user_extra_fields', {})
                user = self.create_user(user_info, extra_fields)

                SocialAccount.objects.create(**{
                    'user': user,
                    self.social_account_field: user_info.get('id')
                })

                self.status_code = status.HTTP_201_CREATED

        token, _ = Token.objects.get_or_create(user=user)
        return user, token


class GoogleAuthView(SocialAuthView):
    platform = 'GOOGLE'
    social_account_field = 'google_id'
    token_url = constants.GOOGLE_TOKEN_URL

    def get(self, request, format=None):
        payload = {
            'response_type': 'code',
            'client_id': self.CLIENT_ID,
            'redirect_uri': self.REDIRECT_URI,
            'scope': (
                'https://www.googleapis.com/auth/userinfo.profile ',
                'https://www.googleapis.com/auth/userinfo.email'
            ),
            'state': json.dumps(request.query_params),
        }
        request = requests.Request('GET', constants.GOOGLE_AUTH_URL, params=payload).prepare()
        return redirect(request.url)

    def get_user_info(self, access_token):
        response = requests.get(constants.GOOGLE_USER_URL, headers={
            'Authorization': 'Bearer %s' % access_token
        })

        if response.status_code != status.HTTP_200_OK:
            raise AuthError(message='Cannot get user info', status=status.HTTP_401_UNAUTHORIZED)

        return response.json()

    def validate_response(self, response):
        if response.status_code != status.HTTP_200_OK:
            raise AuthError(message=response.json(), status=response.status_code)


class FacebookAuthView(SocialAuthView):
    platform = 'FACEBOOK'
    social_account_field = 'facebook_id'
    token_url = constants.FACEBOOK_TOKEN_URL

    def get(self, request, format=None):
        payload = {
            'response_type': 'code',
            'client_id': self.CLIENT_ID,
            'redirect_uri': self.REDIRECT_URI,
            'state': json.dumps(request.query_params),
            'scope': 'email',
        }
        request = requests.Request('GET', constants.FACEBOOK_AUTH_URL, params=payload).prepare()
        return redirect(request.url)

    def get_user_info(self, access_token):
        response = requests.get(constants.FACEBOOK_USER_URL, headers={
            'Authorization': 'Bearer %s' % access_token
        }, params={
            'fields': {'email,first_name,last_name'}
        })

        if response.status_code != status.HTTP_200_OK:
            raise AuthError(message='Cannot get user info.', status=status.HTTP_401_UNAUTHORIZED)

        return response.json()


@api_view(['POST',])
@permission_classes((AllowAny, ))
def password_reset(request):
    serializer = PasswordResetSerializer(data=request.data)
    serializer.is_valid(raise_exception=True)

    form = PasswordResetForm(serializer.validated_data)
    if form.is_valid():
        form.save(
            from_email=settings.CKL_REST_AUTH.get('FROM_EMAIL'),
            email_template_name='registration/password_reset_email.html',
            request=request
        )

    return JsonResponse(request.data, status=status.HTTP_200_OK)



from django.urls import include, path

from . import views

urlpatterns = [
    path('register/', views.RegisterView.as_view(), name='register'),
    path('login/', views.LoginView.as_view(), name='login'),
    path('social/google/', views.GoogleAuthView.as_view(), name='google'),
    path('social/facebook/', views.FacebookAuthView.as_view(), name='facebook'),
    path('password-reset/', views.password_reset, name='password-reset'),
]



from django.urls import include, path

from cklauth.api import urls

urlpatterns = [
    path('api/', include((urls, 'cklauth'), namespace='cklauth')),
    path('', include('django.contrib.auth.urls')),
]



from django.apps import AppConfig
from django.conf import settings


class AuthConfig(AppConfig):
    name = 'cklauth'

    def ready(self):
        setattr(settings, 'CKL_REST_AUTH', {
            'LOGIN_FIELD': 'email',
            'REGISTER_FIELDS': ('username', 'email'),
            'USER_SERIALIZER': 'cklauth.api.v1.serializers.UserSerializer',

            # Override defaults
            **settings.CKL_REST_AUTH,

            # Social defaults
            'GOOGLE': {
                'AUTH_FIELD_GENERATOR': 'cklauth.utils.auth_field_generator',
                'USER_INFO_MAPPING': {
                    'first_name': 'given_name',
                    'last_name': 'family_name',
                    'email': 'email',
                },
                **settings.CKL_REST_AUTH.get('GOOGLE', {}),
            },
            'FACEBOOK': {
                'AUTH_FIELD_GENERATOR': 'cklauth.utils.auth_field_generator',
                'USER_INFO_MAPPING': {
                    'first_name': 'first_name',
                    'last_name': 'last_name',
                    'email': 'email',
                },
                **settings.CKL_REST_AUTH.get('FACEBOOK', {}),
            }
        })



# Google
GOOGLE_TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'
GOOGLE_AUTH_URL = 'https://accounts.google.com/o/oauth2/auth'
GOOGLE_USER_URL = 'https://www.googleapis.com/oauth2/v2/userinfo'
# Facebook
FACEBOOK_TOKEN_URL = 'https://graph.facebook.com/oauth/access_token'
FACEBOOK_AUTH_URL = 'https://www.facebook.com/v2.12/dialog/oauth?'
FACEBOOK_USER_URL = 'https://graph.facebook.com/v2.11/me'



from django.db import models
from django.contrib.auth import settings


class SocialAccount(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        related_name='social_account',
        on_delete=models.CASCADE
    )
    facebook_id = models.CharField(max_length=100, null=True, blank=True)
    google_id = models.CharField(max_length=100, null=True, blank=True)



from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html

from . import models


@admin.register(models.SocialAccount)
class SocialAccountAdmin(admin.ModelAdmin):
    list_display = (
        'user',
    )
    readonly_fields = (
        'user_',
        'facebook_id',
        'google_id',
    )
    exclude = (
        'user',
    )
    search_fields = (
        'user__username',
    )

    def user_(self, instance):
        if not instance.user:
            return '-'
        return format_html('<a href="{0}">{1}</a>'.format(
            reverse('admin:auth_user_change', args=[instance.user.id]),
            instance.user.username or instance.user.id
        ))

    user_.allow_tags = True



from django.contrib.auth import get_user_model, settings
from rest_framework.authentication import TokenAuthentication


User = get_user_model()


class TokenAuthSupportQueryString(TokenAuthentication):
    def authenticate(self, request):
        if 'auth_token' in request.query_params and 'HTTP_AUTHORIZATION' not in request.META:
            return self.authenticate_credentials(request.query_params.get('auth_token'))
        else:
            return super(TokenAuthSupportQueryString, self).authenticate(request)


class EmailOrUsernameModelBackend(object):
    def authenticate(self, request=None, username=None, password=None):
        kwargs = {settings.CKL_REST_AUTH['LOGIN_FIELD']: username}
        try:
            user = User.objects.filter(**kwargs).order_by('id')[0]
            if user.check_password(password):
                return user
        except IndexError:
            return None

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



from os import urandom
from binascii import hexlify

from django.contrib.auth import get_user_model


User = get_user_model()


def auth_field_generator(user_fields, add_suffix=False):
    """
    Generate a login field based on the user first and last name and check if it is in use by any
    user. Recursively adds a hash suffix in case of repeated login field values.

    Args:
        user_fields (dict): user information.
        add_suffix (bool): whether it should calculate and add a random hash to the end of the
            field. Add with the following format `<first-name>-<last-name>-<suffix>`.

    Returns:
        (str) The composed unique login field value.
    """
    base = '-'.join(filter(None, [
        user_fields.get('first_name').lower().replace(" ", "_"),
        user_fields.get('last_name').lower().replace(" ", "_")
    ]))
    suffix = hexlify(urandom(3)) if add_suffix else None
    login_field = '-'.join(filter(None, [base, suffix]))

    if User.objects.filter(**{User.USERNAME_FIELD: base}).exists():
        return auth_field_generator(login_field, add_suffix=True)

    return login_field
